function expe_errgt_geod_curvature(emin, emax)
%EXPE_ERRGT_GEOD_CURVATURE Big experiments on synthetic data.
%
% The following methods are tested:
%  - r1: Richard (measurement error in the image)
%  - r2: Richard (in terms of vector)
%  - p1: Perriolat init
%  - p2: Perriolat refined
%  - s:  Salzmann SOCP
%  - b1: Fitted B-spline
%  - b2: Refined B-spline
%
% The following quantities are measured:
%  - (err_XX) 3D error with the true 3D data points (all algorithms)
%  - (err_gt_XX) 3D error with the ground truth surface (triangular mesh, smooth surface)
%  - (?gauss_XX) Gaussian curvature (only on smooth surface)
%  - (?geod_XX) Geodesic lengths (only on smooth surface)
%
% (c)Florent Brunet

% RIS is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation; either version 3 of the License, or
% (at your option) any later version.
% 
% RIS is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
% or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
% for more details.
% 
% You should have received a copy of the GNU General Public License along
% with this program; if not, write to the Free Software Foundation, Inc.,
% 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
close all;

% Noise amplitude (expressed in pixels in the input image)
noise_amp = 1;
% Noise distribution
noise_type = 'gaussian';

% Noise parameters for the reconstruction algorithms
noise_tpl = noise_amp / 2;
noise_img = noise_amp / 2;

% General experiment parameters
%EXPE_SURF_MIN = 19;
%EXPE_SURF_MAX = 21;
EXPE_SURF_MIN = emin;
EXPE_SURF_MAX = emax;

% Number of point correspondences
n_corresp = 150;

% Number of control points of the B-splines
n_ctrlpts = 12;

% Parameters for B-spline refined reconstruction
sz_inext = 33;
lbd_inext = 50;
lbd_bend = 0.001;

% Parameters for the computation of the Gaussian curvatures
n_gauss_pts = 10000;
EPS_GAUSS = 1.0e-3;

% Parameters for the computation of the pseudo geodesics
n_geod_pts = 10000;

% Number of points taken for approximating the geodesic
n_geod_approx = 200;


for i_expe_surf = EXPE_SURF_MIN:EXPE_SURF_MAX
    time1 = tic;
    
    rand('twister', i_expe_surf); %#ok<RAND>
    randn('state', i_expe_surf); %#ok<RAND>

    % Load a synthetic piece of paper
    % (the .mat file contains the following variables: grid_2d_tpl,
    % grid_2d_img, grid_3d, cam_mat, and sz_grid)
    path = sprintf('data_paper_100/paper%04d.mat', i_expe_surf);
    load(path);
        
    % Extract some point corespondences
    [corresp_2d_tpl corresp_2d_img corresp_3d] = ...
        ris_extract_corresp(n_corresp, grid_2d_tpl, grid_2d_img, grid_3d); %#ok<NODEF>

    % Add noise
    [corresp_2d_tpl corresp_2d_img] = ...
        ris_add_noise(corresp_2d_tpl, corresp_2d_img, noise_type, noise_amp);

    ok_r1 = true;
    ok_r2 = true;

    %% Richard's idea, in term of pixels
    fprintf('r1');
    try
        pts_3d_r1 = ris_richard_errimg( ...
            corresp_2d_tpl, corresp_2d_img, ...
            cam_mat, noise_tpl, noise_img);
    catch excep
        if strcmp(excep.identifier, 'ris_richard_errimg:infeasible')
            ok_r1 = false;
        end
    end

    if ok_r1
        err_r1 = mean(sqrt(sum((corresp_3d - pts_3d_r1).^2)));
    else
        err_r1 = -1;
    end


    %% Richard's idea, in term of vectors
    fprintf(' r2');
    try
        pts_3d_r2 = ris_richard_errvect( ...
            corresp_2d_tpl, corresp_2d_img, ...
            cam_mat, noise_tpl, noise_img/3600);
    catch excep
        if strcmp(excep.identifier, 'ris_richard_errvect:infeasible')
            ok_r2 = false;
        end
    end

    if ok_r2
        err_r2 = mean(sqrt(sum((corresp_3d - pts_3d_r2).^2)));
    else
        err_r2 = -1;
    end


    %% Salzmann's method
    fprintf(' s');
    % Initialization for the Salzmann's method
    umin = min(min(grid_2d_tpl(1,:)), min(corresp_2d_tpl(1,:))) - 1;
    umax = max(max(grid_2d_tpl(1,:)), max(corresp_2d_tpl(1,:))) + 1;
    vmin = min(min(grid_2d_tpl(2,:)), min(corresp_2d_tpl(2,:))) - 1;
    vmax = max(max(grid_2d_tpl(2,:)), max(corresp_2d_tpl(2,:))) + 1;
    npts_mesh = 10;
    [tri tri_x tri_y tri_z] = ris_create_tri_mesh(umin, umax, npts_mesh, vmin, vmax, npts_mesh);

    % Convert the template points into barycentric coordinates
    [b1 b2 b3 ind_tri] = ris_cart_to_bary( ...
        tri, tri_x, tri_y, tri_z, ... 
        corresp_2d_tpl(1,:), corresp_2d_tpl(2,:), ones(size(corresp_2d_tpl(2,:))));

    % Build the inextensibility pairs
    thres_dist = 1.25 * sqrt(2*(max(umax-umin, vmax-vmin) / (npts_mesh - 1))^2);
    [inext_pairs dist_pairs] = ris_make_inext_pairs(tri_x, tri_y, thres_dist);

    % Reconstruction
    % (x,y,z) are the Cartesian coordinates of the mesh _vertices_
    [~, x, y, z] = ris_salzmann_err( ...
                            tri, npts_mesh^2, ...
                            ind_tri, b1, b2, b3, ...
                            corresp_2d_img, ...
                            cam_mat, ...
                            inext_pairs, dist_pairs, ...
                            ones(1,n_corresp), 5);

    [xx yy zz] = bary_to_cart(tri, x, y, z, ind_tri, b1, b2, b3);
    pts_3d_s = [xx ; yy ; zz];

    err_s = mean(sqrt(sum((corresp_3d - pts_3d_s).^2)));
    
    % Some pre-computations for Salzmann's method
    if i_expe_surf == EXPE_SURF_MIN
        % Convert the ground truth template grid points into barycentric coordinates
        [gt_b1 gt_b2 gt_b3 gt_ind_tri] = ris_cart_to_bary( ...
            tri, tri_x, tri_y, tri_z, ... 
            grid_2d_tpl(1,:), grid_2d_tpl(2,:), ones(size(grid_2d_tpl(2,:))));
    end


    [xx yy zz] = bary_to_cart(tri, x, y, z, gt_ind_tri, gt_b1, gt_b2, gt_b3);
    pts_3d_gt_s = [xx ; yy ; zz];
    err_gt_s = mean(sqrt(sum((grid_3d - pts_3d_gt_s).^2)));
    
    
    
    

    %% Perriollat's method (initial upper bounds)
    fprintf(' p1');
    [depth, linkTo, pts_3d_p1, linkAngle] = ...
        ris_perriollat_bounding( ...
            corresp_2d_tpl, corresp_2d_img, ...
            cam_mat, ...
            (noise_tpl+noise_img)); 

    err_p1 = mean(sqrt(sum((corresp_3d - pts_3d_p1).^2)));


    %% Perriollat's method (refined version)
    fprintf(' p2');
    [~, pts_3d_p2] = ...
        ris_perriollat_refined(...
            corresp_2d_tpl, corresp_2d_img, ...
            cam_mat, ...
            depth, linkAngle, linkTo, ...
            0.55*(noise_tpl+noise_img));
        
    err_p2 = mean(sqrt(sum((corresp_3d - pts_3d_p2).^2)));

    
    %% B-spline fitted to R1
    fprintf(' b1');
    EPS = 1.0e-3;
    umin = min(min(grid_2d_tpl(1,:)),min(corresp_2d_tpl(1,:))) - EPS;
    umax = max(max(grid_2d_tpl(1,:)),max(corresp_2d_tpl(1,:))) + EPS;
    nptsu = n_ctrlpts;
    vmin = min(min(grid_2d_tpl(2,:)),min(corresp_2d_tpl(2,:))) - EPS;
    vmax = max(max(grid_2d_tpl(2,:)),max(corresp_2d_tpl(2,:))) + EPS;
    nptsv = n_ctrlpts;
    bbs = bbs_create(umin, umax, nptsu, vmin, vmax, nptsv, 3);

    colmat = bbs_coloc(bbs, corresp_2d_tpl(1,:), corresp_2d_tpl(2,:));
    lbds = 0.1*ones(nptsv-3, nptsu-3);
    bendmat = bbs_bending(bbs, lbds);

    ctrlpts = (colmat'*colmat + bendmat) \ (colmat' * pts_3d_r1');
    ctrlpts = ctrlpts';

    pts_3d_b1 = bbs_eval(bbs, ctrlpts, corresp_2d_tpl(1,:), corresp_2d_tpl(2,:));
    err_b1 = mean(sqrt(sum((corresp_3d - pts_3d_b1).^2)));
    
    pts_3d_gt_b1 = bbs_eval(bbs, ctrlpts, grid_2d_tpl(1,:), grid_2d_tpl(2,:));
    err_gt_b1 = mean(sqrt(sum((grid_3d - pts_3d_gt_b1).^2)));
    
    
    %% Refined B-spline with Jacobian inextensibility
    fprintf(' b2');
    depths_init = sqrt(sum(pts_3d_r1.^2));
    ctrlpts_opt = ...
        ris_jacob_inext_depth_jacob( ...
            corresp_2d_tpl, corresp_2d_img, ...
            cam_mat, ...
            bbs, [depths_init(:) ; vect(ctrlpts')], ...
            sz_inext, lbd_inext, lbd_bend);
        
    pts_3d_b2 = bbs_eval(bbs, ctrlpts_opt, corresp_2d_tpl(1,:), corresp_2d_tpl(2,:));
    err_b2 = mean(sqrt(sum((corresp_3d - pts_3d_b2).^2)));
    
    pts_3d_gt_b2 = bbs_eval(bbs, ctrlpts_opt, grid_2d_tpl(1,:), grid_2d_tpl(2,:));
    err_gt_b2 = mean(sqrt(sum((grid_3d - pts_3d_gt_b2).^2)));
    
    
    %% Gaussian curvatures
    fprintf(' [c]');
    uc = (umax-umin-EPS_GAUSS)*rand(1,n_gauss_pts)+umin+EPS_GAUSS/2;
    vc = (vmax-vmin-EPS_GAUSS)*rand(1,n_gauss_pts)+vmin+EPS_GAUSS/2;
    curv_b1 = gaussian_curvature(bbs, ctrlpts, uc, vc);
    curv_b2 = gaussian_curvature(bbs, ctrlpts_opt, uc, vc);
    
    
    
    %% Pseudo geodesics
    fprintf(' [g]\n');
    ug1 = (umax-umin-EPS_GAUSS)*rand(1,n_geod_pts)+umin+EPS_GAUSS/2;
    vg1 = (vmax-vmin-EPS_GAUSS)*rand(1,n_geod_pts)+vmin+EPS_GAUSS/2;
    ug2 = (umax-umin-EPS_GAUSS)*rand(1,n_geod_pts)+umin+EPS_GAUSS/2;
    vg2 = (vmax-vmin-EPS_GAUSS)*rand(1,n_geod_pts)+vmin+EPS_GAUSS/2;
    
    [eucl, geod_b1] = geodesics(bbs, ctrlpts, ug1, vg1, ug2, vg2, n_geod_approx);
    [~, geod_b2] = geodesics(bbs, ctrlpts_opt, ug1, vg1, ug2, vg2, n_geod_approx);
    
    
    elapsed = toc(time1);

    %% Display and save
    fprintf('(%04d) r1=%f r2=%f s=%f p1=%f p2=%f b1=%f b2=%f\n', i_expe_surf, ...
        err_r1, err_r2, err_s, err_p1, err_p2, err_b1, err_b2);
    fprintf('       gt_s=%f gt_b1=%f gt_b2=%f\n', err_gt_s, err_gt_b1, err_gt_b2);
    fprintf('       curv_b1=%f curv_b2=%f (1/2=%f)\n', mean(curv_b1), mean(curv_b2), mean(curv_b1)/mean(curv_b2));
    fprintf('       geod_b1=%f geod_b2=%f\n', mean((eucl-geod_b1)./eucl), mean((eucl-geod_b2)./eucl));
    fprintf('       [%f s.]\n', elapsed);
    
    
    %% Save
    path_out = sprintf('expe_errgt_geod_curvature/%04d.mat', i_expe_surf);
    save(path_out, ...
        'err_r1', 'err_r2', 'err_s', 'err_p1', 'err_p2', 'err_b1', 'err_b2', ...
        'err_gt_s', 'err_gt_b1', 'err_gt_b2', ...
        'curv_b1', 'curv_b2', ...
        'eucl', 'geod_b1', 'geod_b2');
end




function [x y z] = bary_to_cart(tri, tx, ty, tz, ind_tri, b1, b2, b3)
% Evaluate the 3D position of the data points
for i = 1:numel(b1)
    i1 = tri(ind_tri(i), 1);
    i2 = tri(ind_tri(i), 2);
    i3 = tri(ind_tri(i), 3);
    x(i) = b1(i) * tx(i1) + b2(i) * tx(i2) + b3(i) * tx(i3); %#ok<AGROW>
    y(i) = b1(i) * ty(i1) + b2(i) * ty(i2) + b3(i) * ty(i3); %#ok<AGROW>
    z(i) = b1(i) * tz(i1) + b2(i) * tz(i2) + b3(i) * tz(i3); %#ok<AGROW>
end



function all_kappa = gaussian_curvature(bbs, ctrlpts, u, v)

n_pts = numel(u);

val_u = bbs_eval(bbs, ctrlpts, u, v, 1, 0);
val_v = bbs_eval(bbs, ctrlpts, u, v, 0, 1);
val_uu = bbs_eval(bbs, ctrlpts, u, v, 2, 0);
val_uv = bbs_eval(bbs, ctrlpts, u, v, 1, 1);
val_vv = bbs_eval(bbs, ctrlpts, u, v, 0, 2);

all_kappa = zeros(1,n_pts);
for i = 1:n_pts
    u = val_u(:,i);
    v = val_v(:,i);
    uu = val_uu(:,i);
    uv = val_uv(:,i);
    vv = val_vv(:,i);
    
    % First fundamental form
    E = u'*u;
    F = u'*v;
    G = v'*v;
    
    % Second funcamental form
    n = cross(u,v);
    n = n / norm(n);
    L = uu'*n;
    M = uv'*n;
    N = vv'*n;
    
    all_kappa(i) = (L*N-M^2) / (E*G - F^2);
end




function [all_eucl all_geod] = geodesics(bbs, ctrlpts, u1, v1, u2, v2, n)
n_pts = numel(u1);
all_eucl = sqrt((u1(:)-u2(:)).^2 + (v1(:)-v2(:)).^2)';
all_geod = zeros(1, n_pts);
for i = 1:n_pts
    all_geod(i) = one_geodesic(bbs, ctrlpts, u1(i), v1(i), u2(i), v2(i), n);
end

function geod = one_geodesic(bbs, ctrlpts, u1, v1, u2, v2, n)
all_u = linspace(u1, u2, n);
all_v = linspace(v1, v2, n);
val = bbs_eval(bbs, ctrlpts, all_u, all_v);
tmp = sqrt(sum(([val zeros(3,1)] - [zeros(3,1) val]).^2));
geod = sum(tmp(2:end-1));