function val = ris_cost(corresp_2d_tpl, corresp_2d_img, ...
                        pts_inext_img, ind_pts_inext, dist_inext, lbd_inext, ...
                        mat_intr, bbs, ...
                        bendmat, ...
                        parameters)
%RIS_COST Cost function for "lsqnonlin" for deformable inextensible surface
%reconstruction with our approach.
%
% Input arguments
%
%  - corresp_2d_tpl [2 x n matrix]: 2D points in the template.
%
%  - corresp_2d_img [2 x n matrix]: 2D points in the image.
%
%  - pts_inext_img [2 x m matrix]: 2D points in the image used for the
%     inextensibility constraints. "m" is the total number of
%     "inextensibility points", not the number of "inextensibility
%     constraints".
%
%  - ind_pts_inext [2 x n_constr_inext]: matrix that gives the couple of
%     inextensibility points used to compute the inextensibility
%     constraints. More precisely, there is an inextensibility constraint
%     for all the couples "pts_inext_img(:,ind_pts_inext(1,i))" <->
%     "pts_inext_img(:,ind_pts_inext(2,i))" for i=1,n_constr_inext.
%
%  - dist_inext_img [1 x n_constr_inext vector]: dist_inext_img(i) is the 
%     distance (in the template) between the two points of the i-th
%     inextensibility constraint.
%
%  - lbd_inext [scalar >= 0]: parameter that controls the relative 
%     importance of the inextensibility term.
%
%  - mat_intr [3x3 matrix]: matrix of the intrinsic parameters of the
%     camera.
%
%  - bbs: structure that defines the B-spline (as created with
%     "bbs_create").
%
%  - bendmat [nctrlpts x nctrlpts matrix]: bending matrix (obtained with
%     "bbs_bendmat"). Note that there is no "regularization parameter" 
%     since it is included in the bending matrix itself.
%
% Output arguments
%
%  - val [(n+m+3p)x1 vector with "n" the number of 2D point correspondences,
%     "m" the number of inextensibility constraints, and "p" the number of
%     control points of the warp]: Value of the cost function as required
%     by "lsqnonlin".
%
% (c)2010, Florent Brunet

% RIS is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation; either version 3 of the License, or
% (at your option) any later version.
% 
% RIS is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
% or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
% for more details.
% 
% You should have received a copy of the GNU General Public License along
% with this program; if not, write to the Free Software Foundation, Inc.,
% 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

%% Pre-computation
n_corresp = size(corresp_2d_img, 2);

% Extraction of the diffrent parts of the parameters
[depth ctrlpts ctrlpts_x ctrlpts_y ctrlpts_z] = ...
    ris_split_parameters(parameters, n_corresp);

%% 3D error

% Homogeneous coordinates of the 2D points in the image
corresp_2d_img_hom = [corresp_2d_img ; ones(1, n_corresp)];

% Inverse of the intrinsic matrix of the camera
cam_intr_inv = inv(mat_intr);

% 3D points from the warp
corresp_3d = bbs_eval(bbs, ctrlpts, corresp_2d_tpl(1,:)', corresp_2d_tpl(2,:)');

cost_error3d = vect(corresp_3d ...
    - repmat(depth, 3, 1) .* (cam_intr_inv * corresp_2d_img_hom)); %#ok<MINV>


%% Inextensibility constraints

% 3D warp of the inextensibility points
pts_inext_3d = bbs_eval(bbs, ctrlpts, pts_inext_img(1,:), pts_inext_img(2,:));

% Current distance between the 3D nextensibility points
dist_inext_3d = sqrt(sum((pts_inext_3d(:,ind_pts_inext(1,:)) - pts_inext_3d(:,ind_pts_inext(2,:))).^2));

cost_inext = sqrt(lbd_inext) * (dist_inext_3d - dist_inext)';


%% Bending energy term

% Note that we have to take the squarre root since the expression p'*B*p is
% already the squared bending energy (and the square must not be computed
% in the error vector given to "lsqnonlin").
cost_bending = sqrt([ctrlpts_x * bendmat * ctrlpts_x' ; ...
                     ctrlpts_y * bendmat * ctrlpts_y' ; ...
                     ctrlpts_z * bendmat * ctrlpts_z']);
            
                 
%% Final cost function
val = [cost_error3d ; cost_inext ; cost_bending];

% fprintf('cost=%f  (error3d=%f  inext=%f  bend=%f)\n', sum(val.^2), ...
%     sum(cost_error3d.^2), sum(cost_inext.^2), sum(cost_bending.^2));
