function xopt = ris_gauss_newton(cost, x0)
%RIS_GAUSS_NEWTON Gauss-Newton algorithm for non-linear least squares.
%
% Syntax
%
%  xopt = ris_gauss_newton(cost, x0)
%
% Input arguments
%
%  - cost [function handle]: handle of the cost function. The prototype of
%     this function must be of the form:
%       vec_cost = cost(param)
%     or:
%       [vec_cost jacobian] = cost(param)
%
%  - x0 [vector]: initial solution
%
% Output arguments
%
%  - x_opt [vector]
%
% (c)2010, Florent Brunet

% RIS is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation; either version 3 of the License, or
% (at your option) any later version.
% 
% RIS is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
% or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
% for more details.
% 
% You should have received a copy of the GNU General Public License along
% with this program; if not, write to the Free Software Foundation, Inc.,
% 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

% Parameters of the algorithm
EPS_STEP = 1.0e-6;
EPS_COST = 1.0e-6;
ITER_MAX = 100;
DISPLAY = 2;

% Initialization
old_val_cost = inf;
param = x0;
n_iter = 1;

% Gauss-Newton iterations
lbd = 1;
while true
    [residual jacob JtJ] = cost(param);
    
    % GN step
    while true
        %JtJ = jacob' * jacob;
        diagJtJ = diag(diag(JtJ));
        right = jacob' * residual;
        %delta = - (JtJ + lbd*diagJtJ) \ right;
        delta = - mldivide(JtJ + lbd*diagJtJ, right);
        new_param = param + delta;
        new_residual = cost(new_param);
        new_val_cost = sum(new_residual.^2);
        if (n_iter == 1) || (new_val_cost < val_cost) || (lbd > 1.0e10)
            break;
        end
        fprintf('Increasing lbd (lbd=%f)\n', lbd);
        lbd = 10 * lbd;
    end
    
    param = new_param;
    
    % Values for the stopping criterion
    val_cost = new_val_cost;
    norm_cost = abs(val_cost - old_val_cost);
    norm_step = norm(delta);
    
    % Iteration display
    if DISPLAY >= 2
        fprintf('iter=%d  val_cost=%f  norm_step=%f  evo_cost=%f  lbd=%f\n', ...
            n_iter, val_cost, norm_step, norm_cost, lbd);
    end
    
    % Stopping criterion
    if (norm_cost <= EPS_COST) || (norm_step <= EPS_STEP) || (n_iter >= ITER_MAX)
        break;
    end
    
    old_val_cost = val_cost;
    n_iter = n_iter + 1;
    lbd = max(lbd/10, 1.0e-10);
end

xopt = param;
