function ctrlpts = ris_jacob_inext(corresp_2d_tpl, corresp_2d_img, ...
                                   cam_mat, ...
                                   bbs, ctrlpts_init, ...
                                   sz_inext, lbd_inext)
%RIS_JACOB_INEXT Reconstruction of 3D surface using the "jacobian"
%inextensibility constraints
%
% SYNTAX
%  ctrlpts = ris_jacob_inext(corresp_2d_tpl, corresp_2d_img, ...
%                            cam_mat, ...
%                            bbs, ctrlpts_init, ...
%                            sz_inext, lbd_inext)
%
% DESCRIPTION
%  Uses a discretization of the jacobian inextensibility constraints. The
%  optimization is simply carried with "lsqnonlin".
%
% INPUT ARGUMENTS
%
% OUTPUT ARGUMENTS
%
% (c)2010, Florent Brunet

% RIS is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation; either version 3 of the License, or
% (at your option) any later version.
% 
% RIS is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
% or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
% for more details.
% 
% You should have received a copy of the GNU General Public License along
% with this program; if not, write to the Free Software Foundation, Inc.,
% 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

[u v] = meshgrid(linspace(bbs.umin,bbs.umax,sz_inext), linspace(bbs.vmin,bbs.vmax,sz_inext));
pts_inext_tpl = [u(:)' ; v(:)'];

ctrlpts_opt = lsqnonlin( ...
    @(p)cost( ...
        corresp_2d_tpl, corresp_2d_img, ...
        cam_mat, ...
        pts_inext_tpl, lbd_inext, ...
        bbs, p), ...
    ctrlpts_init(:));

ctrlpts = reshape(ctrlpts_opt, 3, bbs.nptsu*bbs.nptsv);




function val = cost(corresp_2d_tpl, corresp_2d_img, cam_mat, pts_inext_tpl, lbd_inext, bbs, ctrlpts_vect)

n_corresp = size(corresp_2d_tpl, 2);
n_inext = size(pts_inext_tpl, 2);

val = zeros(2*n_corresp + 4*n_inext, 1);

ctrlpts_mat = reshape(ctrlpts_vect, 3, bbs.nptsu*bbs.nptsv);

% Reprojection error
pts_3d = bbs_eval(bbs, ctrlpts_mat, corresp_2d_tpl(1,:), corresp_2d_tpl(2,:));

tmp = cam_mat * [pts_3d ; ones(1, n_corresp)];
repro_2d_img = tmp(1:2,:) ./ repmat(tmp(3,:), 2, 1);

val(1:2*n_corresp) = vect(repro_2d_img - corresp_2d_img);

% Jacobian inextensibility
val_bbs_du = bbs_eval(bbs, ctrlpts_mat, pts_inext_tpl(1,:), pts_inext_tpl(2,:), 1, 0);
val_bbs_dv = bbs_eval(bbs, ctrlpts_mat, pts_inext_tpl(1,:), pts_inext_tpl(2,:), 0, 1);

for i = 1:n_inext
    J = [val_bbs_du(:,i) val_bbs_dv(:,i)];
    val(2*n_corresp+(4*(i-1)+1:4*(i-1)+4), 1) = lbd_inext * vect(J'*J - eye(2));
end

fprintf('cost=%f  (repro=%f, inext=%f)\n', sum(val.^2), ...
    sum(val(1:2*n_corresp).^2), sum(val(2*n_corresp+1:end).^2));
