function ctrlpts = ris_jacob_inext_depth(corresp_2d_tpl, corresp_2d_img, ...
                                         cam_mat, ...
                                         bbs, param_init, ...
                                         sz_inext, lbd_inext)
%RIS_JACOB_INEXT_DEPTH Reconstruction of 3D surface using the "jacobian"
%inextensibility constraints
%
% SYNTAX
%  ctrlpts = ris_jacob_inext_depth(corresp_2d_tpl, corresp_2d_img, ...
%                                  cam_mat, ...
%                                  bbs, param_init, ...
%                                  sz_inext, lbd_inext)
%
% DESCRIPTION
%  Uses a discretization of the jacobian inextensibility constraints. The
%  optimization is simply carried with "lsqnonlin". This version is similar
%  to "ris_jacob_inext" except that here, we parameterize the problem with
%  the control points of 2D-3D warp AND the depth of the data points. In
%  other words, instead of minimizing the reprojection error, we minimize
%  the 3D error.
%
% INPUT ARGUMENTS
%
% OUTPUT ARGUMENTS
%
% (c)2010, Florent Brunet

% RIS is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation; either version 3 of the License, or
% (at your option) any later version.
% 
% RIS is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
% or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
% for more details.
% 
% You should have received a copy of the GNU General Public License along
% with this program; if not, write to the Free Software Foundation, Inc.,
% 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

[u v] = meshgrid(linspace(bbs.umin,bbs.umax,sz_inext), linspace(bbs.vmin,bbs.vmax,sz_inext));
pts_inext_tpl = [u(:)' ; v(:)'];

param_opt = lsqnonlin( ...
    @(p)cost( ...
        corresp_2d_tpl, corresp_2d_img, ...
        cam_mat, ...
        pts_inext_tpl, lbd_inext, ...
        bbs, p), ...
    param_init(:));

n_corresp = size(corresp_2d_tpl, 2);
ctrlpts_vect = param_opt(n_corresp+1:end);
ctrlpts = reshape(ctrlpts_vect, 3, bbs.nptsu*bbs.nptsv);




function val = cost(corresp_2d_tpl, corresp_2d_img, cam_mat, pts_inext_tpl, lbd_inext, bbs, param)

n_corresp = size(corresp_2d_tpl, 2);
n_inext = size(pts_inext_tpl, 2);

val = zeros(3*n_corresp + 4*n_inext, 1);

depths = param(1:n_corresp);
ctrlpts_vect = param(n_corresp+1:end);
ctrlpts_mat = reshape(ctrlpts_vect, 3, bbs.nptsu*bbs.nptsv);



%% 3D error

% 3D points with the warp
pts_3d = bbs_eval(bbs, ctrlpts_mat, corresp_2d_tpl(1,:), corresp_2d_tpl(2,:));

% Retroprojected points
sightlines = normc(cam_mat(1:3,1:3) \ [corresp_2d_img ; ones(1,n_corresp)]);
retro_pts_3d = repmat(depths',3,1) .* sightlines;

val(1:3*n_corresp) = vect((pts_3d - retro_pts_3d).^2);


%% Jacobian inextensibility
val_bbs_du = bbs_eval(bbs, ctrlpts_mat, pts_inext_tpl(1,:), pts_inext_tpl(2,:), 1, 0);
val_bbs_dv = bbs_eval(bbs, ctrlpts_mat, pts_inext_tpl(1,:), pts_inext_tpl(2,:), 0, 1);

for i = 1:n_inext
    J = [val_bbs_du(:,i) val_bbs_dv(:,i)];
    val(3*n_corresp+(4*(i-1)+1:4*(i-1)+4), 1) = lbd_inext * vect(J'*J - eye(2));
end

fprintf('cost=%f  (3D error=%f, inext=%f)\n', sum(val.^2), ...
    sum(val(1:3*n_corresp).^2), sum(val(3*n_corresp+1:end).^2));
