function [depth, linkTo, bounds3D, linkAngle] = ris_perriollat_bounding(pointsTemplate,pointsImage,camera,noiseLevelMax,c,D)
%RIS_PERRIOLLAT_BOUNDING compute the 3D reconstruction of an inextensible 
%surface, giving point correspondances between the current image of the 
%deformed surface and the flat template.
%
% SYNTAX
%  [depth, linkTo, bounds3D, linkAngle] =
%    ris_perriollat_bounding(pointsTemplate, pointsImage, ...
%                            camera, ...
%                            noiseLevelMax, ...
%                            c, D) 
%
% INPUT ARGUMENTS
%   -pointsTemplate: a (2xn) matrix giving the points in the template
%   -pointsImage: a (2xn) matrix giving the points in the current image
%   -camera: a (3x4) matrix giving the projective camera matrix
%   -noiseLevelMax: positive scalar that gives the maximal amount of noise
%      in the template image
%   -c: optional boolean (default: true) that indicate whether the bounds
%      should be refined
%   -D: optional matrix of pre-computed pairwise distances in the template
%
% OUTPUT ARGUMENTS
%   -depth: a (1xn) matrix giving the depth of the points along their 
%      sightline
%   -linkTo: a (1xn) matrix giving the indices of the points from which 
%      the bounds are computed
%   -bounds3D: a (3xn) matrix giving the 3D bound points
%   -linkAngle: a (1xn) matrix giving the angles between the pair of points
%
% NOTES
%  Entirely based on the function "inextensibleTemplateBounding" of Mathieu
%  Perriollat (this is just a refactored version).
%
% (c)2009, Mathieu Perriollat
% (c)2010, Florent Brunet

% RIS is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation; either version 3 of the License, or
% (at your option) any later version.
% 
% RIS is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
% or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
% for more details.
% 
% You should have received a copy of the GNU General Public License along
% with this program; if not, write to the Free Software Foundation, Inc.,
% 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


%% Initial calculation
if ~exist('noiseLevelMax','var')
    noiseLevelMax = 0;
end
nPoints = size(pointsTemplate,2); % the number of points
if ~exist('D','var')
    D = squareform(pdist(pointsTemplate'))+noiseLevelMax;
end
% camera
iM = inv(camera(:,1:3));
C = -iM*camera(:,4); % the camera center
v = normc(iM*[pointsImage;ones(1,nPoints)]); %#ok<MINV> % the sightline directions
% a point Qi along the sightline i is: Qi = \mu*v_i + C

%% Pairwise bound calculation
% pairwise angle computation
alphA = acos(sum(repmat(reshape(v',[1,nPoints,3]),nPoints,1).*permute(repmat(reshape(v',[1,nPoints,3]),nPoints,1),[2 1 3]),3));
alphA(eye(nPoints)==1) = nan;
[depth,linkTo] = min(D./sin(alphA));

if ~exist('c','var')
    c = 1; 
end
while c
    c=0;
    [~,ind] = sort(depth); % to start with the closer point
    for i=ind
        depthTilde = D(i,:)./sin(alphA(i,:));
        % test the distance
        ind_update = depth(i) < D(i,:)./tan(alphA(i,:));
        ind_update(i) = false; % the point bound is not updated
        depthTilde(ind_update) = depth(i).*cos(alphA(i,ind_update)) + sqrt(D(i,ind_update).^2 - depth(i).^2.*sin(alphA(i,ind_update)).^2);
        if max(abs(sqrt((cos(alphA(i,ind_update))*depth(i) - depthTilde(ind_update)).^2 + (sin(alphA(i,ind_update))*depth(i)).^2) - D(i,ind_update)))>1e-10
            pause
        end
        % update the bounds
        [depth,ac] = min([depth;depthTilde]);
        linkTo(ac==2) = i; % the bounds are updated thanks to the ith point
        % continue if changements
        if ~isempty(find(ac==2,1))
            c=1;
        end
    end
end

%% 3D point calculation
bounds3D = repmat(depth,3,1).*v + repmat(C,1,nPoints);

linkAngle = alphA(sub2ind([nPoints nPoints],1:nPoints,linkTo));


