function [depth, Q] = ris_perriollat_refined(pointsTemplate,pointsImage,camera,depthMax,linkAngle,linkTo,noiseLevelMax,D)
%RIS_PERRIOLLAT_REFINED Refined method of M. Perriollat for monocular
%surface reconstruction.
%
% SYNTAX
%  [depth, Q] = ...
%      ris_perriollat_refined(pointsTemplate, pointsImage, ...
%      camera, ...
%      depthMax, linkAngle, linkTo,
%      noiseLevelMax, D)
%
% DESCRIPTION
%  Refined method of M. Perriollat. Simple refactoring of the function
%  initially named "surfaceOptimisation".
%
% (c)2010, Florent Brunet

% RIS is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation; either version 3 of the License, or
% (at your option) any later version.
% 
% RIS is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
% or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
% for more details.
% 
% You should have received a copy of the GNU General Public License along
% with this program; if not, write to the Free Software Foundation, Inc.,
% 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

if ~exist('noiseLevelMax','var')
    noiseLevelMax = 0;
end
% points
nPoints = size(pointsTemplate,2); % the number of points
if ~exist('D','var')
    D = squareform(pdist(pointsTemplate'))+noiseLevelMax;
end

%% sightline computation
% camera
iM = inv(camera(:,1:3));
C = -iM*camera(:,4); % the camera center
v = normc(iM*[pointsImage;ones(1,nPoints)]); % the sightline directions

J1 = eye(nPoints);

J2 = eye(nPoints);
J2(sub2ind([nPoints nPoints],1:nPoints,linkTo)) = 1;

d = D(sub2ind([nPoints,nPoints],1:nPoints,linkTo));
lsqOption = optimset('display' , 'off', 'LargeScale', 'on', 'MaxIter', 500, 'JacobPattern',[J1;J2]);

% initialization computation
ind = ((linkTo(linkTo(1:nPoints)) - (1:nPoints))~=0); 
indNP = find(ind); % the non-pairwise indices
indP = find(~ind); % the pairwise indices

ind = depthMax(linkTo(indNP)) > d(indNP)./tan(linkAngle(indNP));
indNP_NEL = find(ind); % the non-pairwise indices with no exact length
indNP_EL = find(ind); % the non-pairwise indices with exact length

depthInit = depthMax;
depthInit(indP) = d(indP)./(2*sin(linkAngle(indP)/2));
indNP_NEL_A = setdiff(indNP_NEL,linkTo(indNP_NEL)); % the indices that do not influence other points
depthInit(indNP_NEL_A) = cos(linkAngle(indNP_NEL_A)).*depthMax(linkTo(indNP_NEL_A)) + sqrt(d(indNP_NEL_A).^2 - sin(linkAngle(indNP_NEL_A)).^2.*depthMax(linkTo(indNP_NEL_A)).^2);
depthInit = real(depthInit);

depth = lsqnonlin(@(depth)etm(depth,depthMax,linkAngle,linkTo,d),depthInit,zeros(1,nPoints),depthMax,lsqOption);

Q = repmat(depth,3,1).*v + repmat(C,1,nPoints);

function e = etm(depth,depthMax,linkAngle,linkTo,d)
lambda = 10;
e = [(depthMax - depth)  lambda*(sqrt((cos(linkAngle).*depth(linkTo)-depth).^2 + (sin(linkAngle).*depth(linkTo)).^2) - d)];
e = e(:);
