function pts_3d = ris_richard_errvect(corresp_2d_tpl, corresp_2d_img, cam_mat, noise_tpl, noise_img)
%RIS_RICHARD_ERRVECT Inextensible 3D surface reconstruction with the second
%idea of Richard (i.e. error in the image, in terms of vectors).
%
% SYNTAX
%  pts_3d = ris_richard_errvect( ...
%               corresp_2d_tpl, corresp_2d_img, ...
%               cam_mat, ...
%               noise_tpl, noise_img)
%
% INPUT ARGUMENTS
%  - corresp_2d_tpl [2xn matrix]: matrix giving the points in the template.
%  - corresp_2d_img [2xn matrix]: matrix giving the points in the current
%     image.
%  - cam_mat [3x4 matrix]: matrix giving the projective camera matrix.
%  - noise_tpl [scalar]: maximal noise level in the template (used to
%     adjust the distances in the template).
%  - noise_img [scalar]: maximal noise in the image (this one is included
%     in the SOCP reprojection constraints).
%
% OUTPUT ARGUMENTS
%  - pts_3d [3xn matrix]: matrix giving the reconstructed 3D points.
%
% EXCEPTIONS
%  - 'ris_richard_errvect:infeasible': Failure in SeDuMi.
%
% (c)2010, Florent Brunet

% RIS is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation; either version 3 of the License, or
% (at your option) any later version.
% 
% RIS is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
% or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
% for more details.
% 
% You should have received a copy of the GNU General Public License along
% with this program; if not, write to the Free Software Foundation, Inc.,
% 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

% Number of points correspondences
n_corresp = size(corresp_2d_tpl, 2);

% Pairewise distances between the points in the template. 
% "dist_tpl(i,j)" is the euclidean distance between the points
% "corresp_2d_tpl(:,i)" and "corresp_2d_tpl(:,j)".
dist_tpl = squareform(pdist(corresp_2d_tpl'));

% Sightlines
iM = inv(cam_mat(:,1:3));
sightlines = normc(iM * [corresp_2d_img ; ones(1,n_corresp)]); %#ok<MINV> % the sightline directions

% Image error constraints
for k = 1:n_corresp
    tmp = sparse(3, 3*n_corresp);
    tmp(:,3*(k-1)+1:3*(k-1)+3) = sp_mat_prod_vect(sightlines(:,k));
    A{k} = tmp; %#ok<AGROW>
    c{k} = sparse(3, 1); %#ok<AGROW>
    
    tmp = sparse(3*n_corresp,1);
    tmp(3*(k-1)+1:3*(k-1)+3, 1) = noise_img * sightlines(:,k); %#ok<SPRIX>
    b{k} = tmp; %#ok<AGROW>
    d{k} = 0; %#ok<AGROW>
end

% Inextensibility constraints
% k = n_corresp+1;
% ii = [1 2 3 1 2 3];
% v = [ones(1,3) -ones(1,3)];
% for i = 1:n_corresp
%     for j = i+1:n_corresp
%         if dist_tpl(i,j) <= 80
%             jj = [3*(i-1)+1:3*(i-1)+3, 3*(j-1)+1:3*(j-1)+3];
%             A{k} = sparse(ii, jj, v, 3, 3*n_corresp); %#ok<AGROW>
%             c{k} = sparse(3,1); %#ok<AGROW>
%             b{k} = sparse(3*n_corresp,1); %#ok<AGROW>
%             d{k} = dist_tpl(i,j) + noise_tpl; %#ok<AGROW>
%             k = k + 1;
%         end
%     end
% end

%{
% SECOND VERSION
% Here, we just take the immediate neighbors for the inextensibility
% constraints

% We first compute a neighborhood system in the template image
dt = DelaunayTri(corresp_2d_tpl');
tri = dt.Triangulation;
n_facets = size(tri, 1);
adj_lists = cell(n_corresp,1);
for i = 1:n_corresp
    adj_lists{i} = [];
end
for i_facet = 1:n_facets
    i1 = tri(i_facet,1);
    i2 = tri(i_facet,2);
    i3 = tri(i_facet,3);
    
    if i1 > i2, adj_lists{i1}(numel(adj_lists{i1})+1) = i2; end
    if i2 > i3, adj_lists{i2}(numel(adj_lists{i2})+1) = i3; end
    if i3 > i1, adj_lists{i3}(numel(adj_lists{i3})+1) = i1; end
end
for i = 1:n_corresp
    adj_lists{i} = unique(adj_lists{i});
end

k = n_corresp+1;
ii = [1 2 3 1 2 3];
v = [ones(1,3) -ones(1,3)];
for i = 1:n_corresp
    for i_adj = 1:numel(adj_lists{i})
        j = adj_lists{i}(i_adj);
        
        jj = [3*(i-1)+1:3*(i-1)+3, 3*(j-1)+1:3*(j-1)+3];
        A{k} = sparse(ii, jj, v, 3, 3*n_corresp); %#ok<AGROW>
        c{k} = sparse(3,1); %#ok<AGROW>
        b{k} = sparse(3*n_corresp,1); %#ok<AGROW>
        d{k} = dist_tpl(i,j) + noise_tpl; %#ok<AGROW>
        k = k + 1;
    end
end
%}

% THIRD VERSION
% We take the "n_neigh" nearest neighbors
k = n_corresp+1;
n_neigh = 30;
%n_neigh = round(2500 / n_corresp);
ii = [1 2 3 1 2 3];
v = [ones(1,3) -ones(1,3)];
for i = 1:n_corresp
    ln_dist = dist_tpl(i,:);
    [~, ord] = sort(ln_dist);
    ord = ord(ord > i);
    
    for j = ord(1:min(n_neigh,numel(ord)))
        jj = [3*(i-1)+1:3*(i-1)+3, 3*(j-1)+1:3*(j-1)+3];
        A{k} = sparse(ii, jj, v, 3, 3*n_corresp); %#ok<AGROW>
        c{k} = sparse(3,1); %#ok<AGROW>
        b{k} = sparse(3*n_corresp,1); %#ok<AGROW>
        d{k} = dist_tpl(i,j) + noise_tpl; %#ok<AGROW>
        k = k + 1;
    end
end

% Cost function
f = -sightlines(:);

% Optimization
[At, bt, ct, K] = ris_readable_to_sedumi(f, A, c, b, d);
pars.fid = 0;
[~, ys, info] = sedumi(At, bt, ct, K, pars);

if (info.pinf ~= 0) || (info.dinf ~= 0)
    error('ris_richard_errvect:infeasible', 'Infeasible problem');
end

pts_3d = reshape(ys, 3, n_corresp);

function M = mat_prod_vect(u)
%MAT_PROD_VECT Convert a 3x1 vector u into a 3x3 matrix M such that
%prodvect(u,v) = Mu
M = [    0 -u(3)  u(2) ;
      u(3)     0 -u(1) ; 
     -u(2)  u(1)     0 ];

function M = sp_mat_prod_vect(u)
%SP_MAT_PROD_VECT Same as "mat_prod_vect" except that it returns a sparse
%matrix instead of a regular one.
M = sparse(mat_prod_vect(u));
