function [depths, bounds3D] = ris_socp2(pointsTemplate, pointsImage, camera, noiseLevelMax)
%RIS_SOCP2 Inextensible 3D surface reconstruction using SOCP.
%
% Same as "ris_socp" except that "ris_readable_to_sedumi" is used to build
% the SeDuMi parameters of the problem.
%
% Input arguments
%
%  - pointsTemplate [2xn matrix]: matrix giving the points in the template.
%
%  - pointsImage [2xn matrix]: matrix giving the points in the current
%     image.
%
%  - camera [3x4 matrix]: matrix giving the projective camera matrix.
%
%  - noiseLevelMax [scalar, optional (default=0)]: maximal noise level.
%
% Output arguments
%
%  - depths [1xn matrix]: matrix giving the depth of the points along their 
%     sightline.
%
%  - bounds3D [3xn matrix]: matrix giving the 3D bound points.
%
% (c)2010, Florent Brunet

% RIS is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation; either version 3 of the License, or
% (at your option) any later version.
% 
% RIS is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
% or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
% for more details.
% 
% You should have received a copy of the GNU General Public License along
% with this program; if not, write to the Free Software Foundation, Inc.,
% 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

%% Initial computations

% Maximum level of noise
if nargin < 4
    noiseLevelMax = 0;
end

% Number of points correspondences
nPoints = size(pointsTemplate, 2);

% Pairewise distances between the points in the template. 
% "distTemplate(i,j)" is the euclidean distance between the points
% "pointsTemplate(:,i)" and "pointsTemplate(:,j)".
distTemplate = squareform(pdist(pointsTemplate')) + noiseLevelMax;

% Initial computation on the camera
iM = inv(camera(:, 1:3));
camCenter = -iM * camera(:, 4); % the camera center
sightLines = normc(iM * [pointsImage ; ones(1, nPoints)]); %#ok<MINV> % the sightline directions
% (A point Qi along the sightline i is: Qi = \mu*v_i + C)


k = 1;


for i = 1:nPoints
    for j = [1:i-1 i+1:nPoints]
        tmp = sparse(3,nPoints);
        tmp(:,i) = sightLines(:,i);
        tmp(:,j) = -sightLines(:,j);
        A{k} = tmp;
        c{k} = sparse(3,1);
        b{k} = sparse(nPoints,1);
        d{k} = distTemplate(i,j);
        k = k + 1;
    end
end
f = -ones(nPoints,1);

[At, bt, ct, K] = ris_readable_to_sedumi(f, A, c, b, d);

[~, ys] = sedumi(At, bt, ct, K);

depths = ys;

bounds3D = repmat(camCenter, 1, nPoints) + repmat(depths', 3, 1) .* sightLines;
